#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#include "rrdtool.h"
#include "process.h"
#include "parser.h"
#include "debug.h"
#include "string.h"

#define MAX_ARGS	512
#define ARGSTR_LEN	32768

#define RRDTOOL_CMD "/usr/bin/rrdtool"

/*
 * Add new argument to a argument list
 *
 * args		pointer list to arguments
 * argcnt	argument counter
 * argstr	array where the actual arguments are stored
 * idx		index in the argstr where the new argument will be appended
 */
#define add_arg(args, argcnt, argstr, idx, fmt, arg...)	\
	args[argcnt] = argstr + idx;			\
	idx += sprintf(argstr + idx, fmt, ##arg);	\
	argcnt++;					\
	argstr[++idx] = 0

int rrdtool_draw_image(struct rrd_image *image)
{
	int pid;
	char cmd[] = RRDTOOL_CMD;
//	char cmd[] = "echo";
	char *args[512], argstr[ARGSTR_LEN];
	int idx = 0, argcnt = 0, i,j;
	char timestamp[256];
	char tmp[256];
	char tmpfile[256];
	time_t t = time(0);

	pid = do_fork_limited();
	if (pid)
		return pid;

	pr_info("Drawing image %s\n", image->image_filename);

	tmpfile[0] = 0;
	strncat(tmpfile, image->image_filename, sizeof(tmp) - strlen(tmp) - 1);
	strncat(tmpfile, ".tmp", sizeof(tmp) - strlen(tmp) - 1);

	strftime(tmp, 256, "%d.%m.%Y %T (%Z) ", localtime(&t));
	for (i = 0, j = 0; j < 256;) {
		if (tmp[i] == ':') {
			timestamp[j++] = '\\';
		}
		timestamp[j++] = tmp[i++];
		if (!tmp[i])
			break;
	}
	timestamp[j] = 0;


	add_arg(args, argcnt, argstr, idx, RRDTOOL_CMD);
	add_arg(args, argcnt, argstr, idx, "graph");
	add_arg(args, argcnt, argstr, idx, "%s", tmpfile);

	add_arg(args, argcnt, argstr, idx, "--start");
	add_arg(args, argcnt, argstr, idx, "%s", image->timestart);
	add_arg(args, argcnt, argstr, idx, "--end");
	add_arg(args, argcnt, argstr, idx, "%s", image->timeend);
	add_arg(args, argcnt, argstr, idx, "--width");
	add_arg(args, argcnt, argstr, idx, "%d", image->width);
	add_arg(args, argcnt, argstr, idx, "--height");
	add_arg(args, argcnt, argstr, idx, "%d", image->height);
	add_arg(args, argcnt, argstr, idx, "--imgformat");
	add_arg(args, argcnt, argstr, idx, "%s", image->imageformat);

	for (i = 0; image->options[i]; i++) {
		add_arg(args, argcnt, argstr, idx, "%s", image->options[i]);
	}

	for (i = 0; image->text[i]; i++) {
		args[argcnt++] = (char *)image->text[i];
	}

	add_arg(args, argcnt, argstr, idx, "COMMENT:Last update %s\\c", timestamp);

	args[argcnt] = 0;

	pid = run(cmd, args);
	harvest_zombies(pid);

	rename(tmpfile, image->image_filename);

	exit(0);
}

int rrdtool_draw_images(struct rrd_image **image)
{
	int i;
	for (i = 0; image[i]; i++)
		rrdtool_draw_image(image[i]);

	return 0;
}

static int sanitize_rrd_update_data(char *data)
{
	char clean_data[RRD_DATA_MAX_LEN];
	int entries = 0;
	int minus;
	char *src, *end, *cln;

	data[RRD_DATA_MAX_LEN - 1] = 0;
	src = data;
	cln = clean_data;

	/*
	 * Copy a legit floating point number to clean_data buffer
	 * starting from *src and ending to next ':'. If no legit
	 * number could be found, put a 'U' there instead to make
	 * rrdtool to understand this datapoint is undefined.
	 */

	while (src < data + RRD_DATA_MAX_LEN && *src) {
		minus = 0;

		/* skip any non_numbers but not ':' */
		while (*src && !isdigit(*src) && *src != '-' && *src != ':')
			src++;

		if (*src == '-') {
			src++;
			minus = 1;
		}

		/* Now find the end of the number */
		end = skip_numbers(src);

		/* Floating point numberrs may have a dot with more numbers */
		if (*end == '.') {
			end++;
			end = skip_numbers(end);
		}

		/*
		 * Now we have gone past the number, there should be a
		 * colon or zero byte. If src == end, there was no
		 * number and the entry is undefined instead.
		 */
		if ((*end == ':' || !*end) && src != end) {
			if (minus) {
				*cln = '-';
				cln++;
			}

			/*
			 * Copy the legit number and start copying the
			 * next one
			 */
			for (; src <= end; src++, cln++)
				*cln = *src;

			goto next;
		}

		/* Skip over whatever junk there might be */
		while (*end != ':' && *end)
			end++;

		/* Mark the entry as undefined */
		*cln = 'U';
		cln++;
		*cln = ':';
		cln++;
	next:
		end++;
		src = end;
		entries++;
	}

	/*
	 * If last entry was undefined, we need to remove the extra
	 * colon at the end
	 */
	if (*(cln - 1) == ':')
		cln--;
	*cln = '\0';

	strncpy(data, clean_data, RRD_DATA_MAX_LEN);
	return entries;
}

int rrdtool_update_data(struct rrd_database *rrd)
{
	int pid;
	char data[RRD_DATA_MAX_LEN + 3]; /* 3 == "N:" + NULL termination */
	char cmd[] = RRDTOOL_CMD;
//	char cmd[] = "echo";
	char *const cmdline[] = {
		RRDTOOL_CMD,
		"update",
		(char *const)rrd->filename,
		data,
		0
	};
	int l;

	rrd->last_update = time(0);
	if (do_fork())
		return 0;

	l = sprintf(data, "N:");

	if (rrd->parse) {
		rrd->parse(data + l, rrd->parser_data);
		data[RRD_DATA_MAX_LEN + 2] = '\0';

		pr_info("Data: %s\n", data);

		sanitize_rrd_update_data(data + l);
		pid = run(cmd, cmdline);
		harvest_zombies(pid);
	}

	if (rrd->images)
		rrdtool_draw_images(rrd->images);

	while (harvest_zombies(0));
	exit(0);
}

/*
 * Walk through the database list and return the first database which
 * last update is too far in past
 */
struct rrd_database *get_outdated_db(struct rrd_database **dblist)
{
	int i;
	time_t now = time(0);

	for (i = 0; dblist[i]; i++) {
		if ((dblist[i]->last_update + dblist[i]->interval) - now <= 0)
			return dblist[i];
	}

	/* Nothing to update this time, return null */
	return NULL;
}

/*
 * See how long we may sleep until it is required to run an update
 * again
 */
int get_next_update(struct rrd_database **dblist, const char **name)
{
	int i, sleeptime = 0, diff;
	time_t now = time(0);

	for (i = 0; dblist[i]; i++) {
		diff = dblist[i]->last_update + dblist[i]->interval - now;
		if (!sleeptime) {
			sleeptime = diff;
			*name = dblist[i]->name;
		}
		if (sleeptime > diff) {
			sleeptime = diff;
			*name = dblist[i]->name;
		}
		if (sleeptime <= 0)
			return 0;
	}

	return sleeptime;
}

static int database_exists(struct rrd_database *db)
{
	struct stat s;

	/* If the filename exists, stat will return zero */
	if (db->filename)
		return !stat(db->filename, &s);

	return 0;
}

static int create_database(struct rrd_database *db)
{
	char cmd[] = RRDTOOL_CMD;
//	char cmd[] = "echo";
	char *args[512], argstr[ARGSTR_LEN];
	int idx = 0, argcnt = 0;
	int child, i;

	if (!db->filename) {
		pr_err("Database %s missing database filename\n", db->name);
		return -1;
	}

	if (!db->sources || !db->archives) {
		pr_err("Cannot create db \"%s\", insufficient source data\n",
			db->filename);
		return -1;
	}

	add_arg(args, argcnt, argstr, idx, RRDTOOL_CMD);
	add_arg(args, argcnt, argstr, idx, "create");
	add_arg(args, argcnt, argstr, idx, "%s", db->filename);
	add_arg(args, argcnt, argstr, idx, "--step");
	add_arg(args, argcnt, argstr, idx, "%d", db->interval);

	for (i = 0; db->sources[i].type; i++) {
		add_arg(args, argcnt, argstr, idx, "DS:%s:%s:%d:%f:%f",
			db->sources[i].name,
			db->sources[i].type,
			db->sources[i].heartbeat,
			db->sources[i].min,
			db->sources[i].max);
	}

	for (i = 0; db->archives[i].type; i++) {
		add_arg(args, argcnt, argstr, idx, "RRA:%s:%f:%d:%d",
			db->archives[i].type,
			db->archives[i].xff,
			db->archives[i].steps,
			db->archives[i].rows);
	}

	child = run(cmd, args);

	harvest_zombies(child);

	return 0;
}

int rrdtool_create_missing_databases(struct rrd_database *dbs[])
{
	struct rrd_database *db;
	int i, ret = 0;

	for (i = 0, db = dbs[i]; db; i++, db = dbs[i]) {
		if (database_exists(db)) {
			pr_info("database %s found\n", db->filename);
			continue;
		}
		pr_info("Database %s missing, creating\n", db->filename);
		ret |= create_database(db);
	}

	return ret;
}
